# coding: utf-8
import h5py
import numpy as np
from getdist import MCSamples
import matplotlib.pyplot as plt
import matplotlib

# Default Plot labels for datasets, can be overwritten with dictionary passed to 'get_MCSamples_from_gambit'
glob_labels_dict = {
    'H0':r'$H_{0}$',
    'ln10A_s':r'$\ln{10^{10} A_{s}}$',
    'omega_b':r'$\Omega_\mathrm{b} h^2$',
    'omega_cdm':r'$\Omega_\mathrm{cdm} h^2$',
    'n_s':r'$n_{s}$',
    'tau_reio':r'$\tau_\mathrm{reio}$',
    'dNur_CMB':r'$\Delta N_\mathrm{ur}$',
    'r_CMB':r'$r_\nu$',
    'dNeff':r'$\Delta N_\mathrm{eff}$',
    'Neff':r'$N_\mathrm{eff}$',
    'log_dNeff':r'$\log_{10}\,\Delta N_\mathrm{eff}$',
    'Smu':r'$\Sigma\,m_\nu\,\mathrm{[eV]}$',
    'log_Smu':r'$\log_{10}\,\Sigma\,m_\nu$',
    'M':r'$M_\mathrm{abs}$',
    'mNu_light':r'$m_{\nu,\mathrm{light}}\,\mathrm{[eV]}$',
    'dmNu21':r'$\Delta m^{2}_{21}$',
    'dmNu3l':r'$\Delta m^{2}_{3\ell}$',
    'Sigma8':r'$\Sigma_8$',
    'Omega_m':r'$\Omega_\mathrm{m}$',
}


# Bounds for parameters which will hit the left and/or right edge of the parameter space. If no bound applies, choose None
glob_bounds_dict = {
    'tau_reio':(0.004,None),
    'Smu':(0,None),
    'mNu_light':(0.0,None),
}

# Main function to create a getdist.MCSamples instance out of the data in the hdf5 file
def get_MCSamples_from_gambit(gambit_hdf5_file,**kwargs):
    # Open the hdf5 file in read-only mode
    read_data = h5py.File(gambit_hdf5_file,'r')

    # What is the group of the hdf5file? 
    group = kwargs.pop('group','data')

    # Any additonal datasets  requested,which are not included by default?
    # (By default, only the primary_parameters, the total LogLike,
    #  and the posterior weight will be included)
    additional_keys = kwargs.pop('additional_keys',None)

    # Which parameters should be plotted in log10 space?
    logspace_list = kwargs.pop('logSpace',list())

    # Are they any specific labels used to superceed the ones in glob_labels_dict?
    _labels_dict = kwargs.pop('labels_dict', dict())
    labels_dict = dict()
    labels_dict.update(glob_labels_dict)
    labels_dict.update(_labels_dict)

    # Are there any updated paramter bounds
    _bounds_dict = kwargs.pop('bounds_dict', dict())
    bounds_dict = dict()
    bounds_dict.update(glob_bounds_dict)
    bounds_dict.update(_bounds_dict)

    # Which sampler was in use? -> How is the posterior dataset called?
    sampler = kwargs.pop('sampler','multinest')
    post_label = {'twalk':'mult', 'multinest':'Posterior', 'polychord':'Posterior'}.get(sampler)
    if post_label not in read_data.get(group).keys():
        raise Exception("I cannot find the dataset '{:s}'".format(post_label))

    total_mask = None
    # If we have twalk samples, throw out all rejected samples (chain == -1)
    if sampler == 'twalk':
        total_mask = (np.array( read_data.get('data/chain'), dtype=read_data.get('data/chain').dtype) != -1)
        total_mask = np.logical_and(total_mask, np.array( read_data.get('data/chain_isvalid'), dtype=np.bool))

    tmp_dict = dict()
    for key in read_data.get(group).keys():
        if key.endswith('isvalid'):
        # There are two arrays for every dataset ('NAME' and 'NAME_isvalid') containing the data and a 
        # flag if the data is valid (sampling point was not invalidated etc.)
        # Search only for the keys with 'isvalid' in them
            key_of_new_data = None
            maskkey = '{:s}/{:s}'.format(group,key)
            datakey = '{:s}/{:s}'.format(group,key.replace("_isvalid",""))
            # Strip the _isvalid suffix
            key = key.replace("_isvalid","")

            # Only search for Gambit model parameters (They contain 'primary_parameters')
            # and use their name (after ::) as name of the dataset
            if 'primary_parameters' in key:
                key_of_new_data = datakey.split(':')[-1]
            # Include the total LogLike (key must match LogLike_isvalid)
            elif key == 'LogLike':
                key_of_new_data = 'LogLike'
            # Include the posterior weight (which has diferent names, depending on the scanner)
            elif key == post_label:
                key_of_new_data = 'Posterior'
            # Skip everything else
            else:
                pass

            # Include extra parameter, which are not include otherwise
            if additional_keys is not None:
                for add_key, add_value in additional_keys.items():
                    if add_key in datakey:
                        key_of_new_data = add_value

            if key_of_new_data is not None:
                # get the data and the mask
                tmp_dat = np.array( read_data[datakey],dtype=read_data[datakey].dtype)
                # Go to log10 space if requested
                if key_of_new_data in logspace_list:
                    m = (tmp_dat > 0.0)
                    tmp_dat[m] = np.log10(tmp_dat[m])
                    tmp_dat[~m] = np.nan
                # Get the original mask (from "_isvalid")
                # but only if it is an observable or likelihood
                # and neither a primary parameter nor the posterior dataset
                if not ('primary_parameters' in key or key == post_label):
                    tmp_mask = np.array( read_data[maskkey], dtype=np.bool)
                else:
                    tmp_mask = np.ones_like(tmp_dat).astype(np.bool)
                # Mask nonfinite entries (Can occur during log10 transformation)
                tmp_mask = np.logical_and(tmp_mask, np.isfinite(tmp_dat))

                # If the dataset has some hard bounds, mask all entries which are not within this range
                boundmask = np.ones_like(tmp_dat).astype(np.bool)
                if key_of_new_data in bounds_dict:
                    lo, hi = bounds_dict.get(key_of_new_data)
                    if lo is None:
                        lo = -np.inf
                    if hi is None:
                        hi = np.inf
                    boundmask = np.logical_and( (tmp_dat >= lo), (tmp_dat <= hi) )
                tmp_mask = np.logical_and(tmp_mask, boundmask)

                # Finally combine the masks (element-wise AND)
                if total_mask is None:
                    total_mask = tmp_mask
                else:
                    total_mask = np.logical_and(tmp_mask,total_mask)

                # Put the dataset into the dictionary
                tmp_dict.update({key_of_new_data:tmp_dat})

    # We can close the hdf5 file now as the dataset dictionary is filled
    read_data.close()

    data_dict = dict()

    # Loop through tmp_dict and add the item to data_dict if they are varying
    # Also add LogLike (as -Loglike) and Posterior
    for key, values in tmp_dict.items():
        if key != 'LogLike' and key != 'Posterior':
            if np.sqrt(values[total_mask].var())/abs(values[total_mask].mean()) > 1e-9:
                # only include data if they vary. Exclude fixed parameters.
                data_dict.update({key:values[total_mask]})
        elif key == 'LogLike':
            # invert the logLike array. 
            data_dict.update({key:-values[total_mask]})
        else: # (key == 'Posterior')
            data_dict.update({key:values[total_mask]})

    # Get the parameter names
    names = list(data_dict.keys())
    names.remove('Posterior')
    names.remove('LogLike')

    # Get ndim
    ndim = len(names)

    # Get sample size
    nsamples = data_dict.get(names[0]).shape[-1]

    # Prepare empty arrays for data and labels
    # and an empty dict for rnages
    data = np.zeros((nsamples,ndim), dtype='f8')
    labels = np.empty((ndim), dtype='U128')
    ranges = dict()

    # Loop through all names and set data, label, and ranges
    for idx, key in enumerate(names):
        data[:,idx] = data_dict.get(key)
        # Try to get the label from the global dictionary.
        # If not applicable, wrap the parameter name into \mathrm{} 
        labels[idx] = labels_dict.get(key,(r'$\mathrm{'+key+'}$').replace('_','\_'))
        if key in bounds_dict:
            ranges.update({key:bounds_dict.get(key,(None,None))})

    # Hand everything over to getdist and get the MCsamples instance
    samples = MCSamples(samples=data,
                        weights=data_dict.get('Posterior').astype('f8')/data_dict.get('Posterior').sum(), 
                        loglikes=data_dict.get('LogLike'),
                        names=names,
                        labels=labels,
                        ranges=ranges,
                        **kwargs)

    return samples

# Function to get x and f(x) whwere f(x) is the 1D pdf of x
def get_1D_posterior(mcsamples,param,scale=1.0):
    settings = {
        'smooth_scale_1D':mcsamples.smooth_scale_1D,
        'boundary_correction_order':mcsamples.boundary_correction_order,
        'mult_bias_correction_order':mcsamples.mult_bias_correction_order,
        'fine_bins':mcsamples.fine_bins,
        'num_bins':mcsamples.num_bins,
    }

    d = mcsamples.get1DDensity(param,**settings)
    low, high = d.view_ranges
    x = d.x
    y = d.P
    m = np.where((x >= low) & (x <= high))
    return (scale*x[m], y[m])

# Function to return the paths (x,y) coordinates for a specific confidence level
def get_2D_contour_level(mcsamples,level,param1,param2,scale1=1.0,scale2=1.0):
    settings = {
        'smooth_scale_2D':mcsamples.smooth_scale_2D,
        'boundary_correction_order':mcsamples.boundary_correction_order,
        'mult_bias_correction_order':mcsamples.mult_bias_correction_order,
        'fine_bins_2D':mcsamples.fine_bins_2D,
        'num_bins_2D':mcsamples.num_bins_2D,
    }

    d = mcsamples.get2DDensity(param1, param2,**settings)
    c = d.getContourLevels([level])
    x = d.x
    y = d.y
    z = d.P

    cnt = plt.contour(x,y,z,[c[0]])
    x1, y1 = cnt.collections[0].get_paths()[0].vertices.T
    return (scale1*x1, scale2*y1)

# Retrun the whole grid x,y, f(x,y) where f(x,y) is the 2D pdf of the x and y marginalised over all other parameters
# also returns the arry c which links a certain confidence level to the respective value of the isocontur.
def get_2D_contour(mcsamples,param1,param2,level=None,scale1=1.0,scale2=1.0):
    settings = {
        'smooth_scale_2D':mcsamples.smooth_scale_2D,
        'boundary_correction_order':mcsamples.boundary_correction_order,
        'mult_bias_correction_order':mcsamples.mult_bias_correction_order,
        'fine_bins_2D':mcsamples.fine_bins_2D,
        'num_bins_2D':mcsamples.num_bins_2D,
    }

    d = mcsamples.get2DDensity(param1, param2,**settings)
    if level is None:
        level = [0.68,0.95]
    level = np.asarray(level)
    c = d.getContourLevels(level)
    c = c[c.argsort()]
    c = np.append(c,1.0)
    x = d.x
    y = d.y
    z = d.P

    return (scale1*x, scale2*y, z,  c)

# Function to provide a paler version of a given color
# Usually used in 2D plots to show th 68%-95% range
# Create paler versions of a original color (useful for contour plots)
# use rounds if you want an even paler version by make_paler ( make_paler_( ... ) )
def make_paler(original_color,rounds=1):
    from matplotlib.colors import rgb2hex, ColorConverter
    pale_factor = .55
    color = ColorConverter.to_rgb(original_color)
    for _ in range(rounds):
        color = [c * (1 - pale_factor) + pale_factor for c in color]
    return "{:s}".format(rgb2hex(color))

# Just for convenience we provide the paler versions of all ten colors of the standard color cycle of matplotlib
C0pale = make_paler('C0')
C1pale = make_paler('C1')
C2pale = make_paler('C2')
C3pale = make_paler('C3')
C4pale = make_paler('C4')
C5pale = make_paler('C5')
C6pale = make_paler('C6')
C7pale = make_paler('C7')
C8pale = make_paler('C8')
C9pale = make_paler('C9')
